!> @file module_interface.f90
!------------------------------------------------------------------------------!
! This file is part of PALM.
!
! PALM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2019 Leibniz Universitaet Hannover
!------------------------------------------------------------------------------!
!
! Current revisions:
! -----------------
! 
! 
! Former revisions:
! -----------------
! $Id: module_interface.f90 3687 2019-01-22 10:42:06Z knoop $
! Some interface calls moved to module_interface + cleanup
! 
! 3684 2019-01-20 20:20:58Z knoop
! Bugfix: made unit intend INOUT
! 
! 3650 2019-01-04 13:01:33Z kanani
! Add restart routines for biometeorology
! 
! 3649 2019-01-02 16:52:21Z suehring
! Initialize strings, in order to avoid compiler warnings for non-initialized
! characters with intent(out) attribute
! 
! 3648 2019-01-02 16:35:46Z suehring
! Rename subroutines for surface-data output
! 
! 3641 2018-12-23 22:10:01Z knoop
! Initial implementation of the PALM module interface
!
! 
! Description:
! ------------
!> This is the interface between the PALM model core and all its modules.
!>
!------------------------------------------------------------------------------!
MODULE module_interface

   USE indices,                                                                &
       ONLY:  nbgp, nxl, nxlg, nxr, nxrg, nys, nysg, nyn, nyng, nzb, nzt

   USE kinds

!- load module-specific control parameters.
!- ToDo: move all of them to respective module or a dedicated central module

   USE control_parameters,                                                     &
       ONLY:  biometeorology,                                                  &
              air_chemistry,                                                   &
              indoor_model,                                                    &
              land_surface,                                                    &
              large_scale_forcing,                                             &
              nesting_offline,                                                 &
              nudging,                                                         &
              ocean_mode,                                                      &
              plant_canopy,                                                    &
              salsa,                                                           &
              surface_output,                                                  &
              syn_turb_gen,                                                    &
              urban_surface,                                                   &
              virtual_flight,                                                  &
              virtual_measurement,                                             &
              wind_turbine

!- load interface routines of all PALM modules

   USE biometeorology_mod,                                                     &
       ONLY:  bio_parin,                                                       &
              bio_check_parameters,                                            &
              bio_check_data_output,                                           &
              bio_init,                                                        &
              bio_header,                                                      &
              bio_3d_data_averaging,                                           &
              bio_data_output_2d,                                              &
              bio_data_output_3d,                                              &
              bio_rrd_global,                                                  &
              bio_rrd_local,                                                   &
              bio_wrd_global,                                                  &
              bio_wrd_local

   USE bulk_cloud_model_mod,                                                   &
       ONLY:  bulk_cloud_model,                                                &
              bcm_parin,                                                       &
              bcm_check_parameters,                                            &
              bcm_check_data_output_pr,                                        &
              bcm_check_data_output,                                           &
              bcm_init_arrays,                                                 &
              bcm_init,                                                        &
              bcm_header,                                                      &
              bcm_swap_timelevel,                                              &
              bcm_3d_data_averaging,                                           &
              bcm_data_output_2d,                                              &
              bcm_data_output_3d,                                              &
              bcm_rrd_global,                                                  &
              bcm_wrd_global,                                                  &
              bcm_rrd_local,                                                   &
              bcm_wrd_local

   USE chemistry_model_mod,                                                    &
       ONLY:  chem_parin,                                                      &
              chem_check_parameters,                                           &
              chem_check_data_output_pr,                                       &
              chem_check_data_output,                                          &
              chem_init_arrays,                                                &
              chem_init,                                                       &
              chem_header,                                                     &
              chem_swap_timelevel,                                             &
              chem_3d_data_averaging,                                          &
              chem_data_output_2d,                                             &
              chem_data_output_3d,                                             &
              chem_statistics,                                                 &
              chem_rrd_local,                                                  &
              chem_wrd_local

   USE flight_mod,                                                             &
       ONLY:  flight_parin,                                                    &
              flight_header,                                                   &
              flight_init,                                                     &
              flight_rrd_global,                                               &
              flight_wrd_global

   USE gust_mod,                                                               &
       ONLY:  gust_module_enabled,                                             &
              gust_parin,                                                      &
              gust_check_parameters,                                           &
              gust_check_data_output_pr,                                       &
              gust_check_data_output,                                          &
              gust_init_arrays,                                                &
              gust_init,                                                       &
              gust_header,                                                     &
              gust_actions,                                                    &
              gust_swap_timelevel,                                             &
              gust_3d_data_averaging,                                          &
              gust_data_output_2d,                                             &
              gust_data_output_3d,                                             &
              gust_statistics,                                                 &
              gust_rrd_global,                                                 &
              gust_wrd_global,                                                 &
              gust_rrd_local,                                                  &
              gust_wrd_local

   USE indoor_model_mod,                                                       &
       ONLY:  im_parin,                                                        &
              im_init

   USE land_surface_model_mod,                                                 &
       ONLY:  lsm_parin,                                                       &
              lsm_check_parameters,                                            &
              lsm_check_data_output_pr,                                        &
              lsm_check_data_output,                                           &
              lsm_init_arrays,                                                 &
              lsm_init,                                                        &
              lsm_header,                                                      &
              lsm_swap_timelevel,                                              &
              lsm_3d_data_averaging,                                           &
              lsm_data_output_2d,                                              &
              lsm_rrd_local,                                                   &
              lsm_wrd_local

   USE lsf_nudging_mod,                                                        &
       ONLY:  lsf_nudging_check_parameters,                                    &
              lsf_nudging_check_data_output_pr,                                &
              lsf_init,                                                        &
              nudge_init,                                                      &
              lsf_nudging_header

   USE multi_agent_system_mod,                                                 &
       ONLY:  mas_parin

   USE nesting_offl_mod,                                                       &
       ONLY:  nesting_offl_parin,                                              &
              nesting_offl_check_parameters,                                   &
              nesting_offl_header

   USE ocean_mod,                                                              &
       ONLY:  ocean_parin,                                                     &
              ocean_check_parameters,                                          &
              ocean_check_data_output_pr,                                      &
              ocean_check_data_output,                                         &
              ocean_init_arrays,                                               &
              ocean_init,                                                      &
              ocean_header,                                                    &
              ocean_swap_timelevel,                                            &
              ocean_3d_data_averaging,                                         &
              ocean_data_output_2d,                                            &
              ocean_data_output_3d,                                            &
              ocean_rrd_global,                                                &
              ocean_wrd_global,                                                &
              ocean_rrd_local,                                                 &
              ocean_wrd_local

   USE plant_canopy_model_mod,                                                 &
        ONLY: pcm_parin,                                                       &
              pcm_check_parameters,                                            &
              pcm_check_data_output,                                           &
              pcm_init,                                                        &
              pcm_header,                                                      &
              pcm_data_output_3d

   USE radiation_model_mod,                                                    &
       ONLY:  radiation,                                                       &
              radiation_parin,                                                 &
              radiation_check_parameters,                                      &
              radiation_check_data_output_pr,                                  &
              radiation_check_data_output,                                     &
              radiation_init,                                                  &
              radiation_header,                                                &
              radiation_3d_data_averaging,                                     &
              radiation_data_output_2d,                                        &
              radiation_data_output_3d,                                        &
              radiation_rrd_local,                                             &
              radiation_wrd_local

   USE salsa_mod,                                                              &
       ONLY:  salsa_parin,                                                     &
              salsa_check_parameters,                                          &
              salsa_check_data_output,                                         &
              salsa_init_arrays,                                               &
              salsa_init,                                                      &
              salsa_header,                                                    &
              salsa_swap_timelevel,                                            &
              salsa_3d_data_averaging,                                         &
              salsa_data_output_2d,                                            &
              salsa_data_output_3d,                                            &
              salsa_rrd_local,                                                 &
              salsa_wrd_local

   USE spectra_mod,                                                            &
       ONLY:  calculate_spectra,                                               &
              spectra_parin,                                                   &
              spectra_check_parameters,                                        &
              spectra_header

   USE surface_data_output_mod,                                                &
       ONLY:  surface_data_output_parin,                                       &
              surface_data_output_check_parameters

   USE synthetic_turbulence_generator_mod,                                     &
       ONLY:  stg_parin,                                                       &
              stg_check_parameters,                                            &
              stg_header,                                                      &
              stg_rrd_global,                                                  &
              stg_wrd_global

   USE urban_surface_mod,                                                      &
       ONLY:  usm_parin,                                                       &
              usm_check_parameters,                                            &
              usm_check_data_output,                                           &
              usm_init_arrays,                                                 &
              usm_init,                                                        &
              usm_swap_timelevel,                                              &
              usm_3d_data_averaging,                                           &
              usm_rrd_local,                                                   &
              usm_wrd_local

   USE user,                                                                   &
       ONLY:  user_parin,                                                      &
              user_check_parameters,                                           &
              user_check_data_output_pr,                                       &
              user_check_data_output,                                          &
              user_init,                                                       &
              user_header,                                                     &
              user_actions,                                                    &
              user_3d_data_averaging,                                          &
              user_data_output_2d,                                             &
              user_data_output_3d,                                             &
              user_statistics,                                                 &
              user_rrd_global,                                                 &
              user_rrd_local,                                                  &
              user_wrd_global,                                                 &
              user_wrd_local,                                                  &
              user_last_actions

   USE virtual_measurement_mod,                                                &
       ONLY:  vm_parin,                                                        &
              vm_init

   USE wind_turbine_model_mod,                                                 &
       ONLY:  wtm_parin,                                                       &
              wtm_check_parameters,                                            &
              wtm_init_arrays,                                                 &
              wtm_init,                                                        &
              wtm_rrd_global,                                                  &
              wtm_wrd_global

   IMPLICIT NONE

   PRIVATE

!
!- Public functions
   PUBLIC                                                                      &
      module_interface_parin,                                                  &
      module_interface_check_parameters,                                       &
      module_interface_check_data_output_pr,                                   &
      module_interface_check_data_output,                                      &
      module_interface_init_masks,                                             &
      module_interface_define_netcdf_grid,                                     &
      module_interface_init_arrays,                                            &
      module_interface_init,                                                   &
      module_interface_header,                                                 &
      module_interface_actions,                                                &
      module_interface_swap_timelevel,                                         &
      module_interface_3d_data_averaging,                                      &
      module_interface_data_output_2d,                                         &
      module_interface_data_output_3d,                                         &
      module_interface_statistics,                                             &
      module_interface_rrd_global,                                             &
      module_interface_wrd_global,                                             &
      module_interface_rrd_local,                                              &
      module_interface_wrd_local,                                              &
      module_interface_last_actions


   INTERFACE module_interface_parin
      MODULE PROCEDURE module_interface_parin
   END INTERFACE module_interface_parin

   INTERFACE module_interface_check_parameters
      MODULE PROCEDURE module_interface_check_parameters
   END INTERFACE module_interface_check_parameters

   INTERFACE module_interface_check_data_output_pr
      MODULE PROCEDURE module_interface_check_data_output_pr
   END INTERFACE module_interface_check_data_output_pr

   INTERFACE module_interface_check_data_output
      MODULE PROCEDURE module_interface_check_data_output
   END INTERFACE module_interface_check_data_output

   INTERFACE module_interface_init_masks
      MODULE PROCEDURE module_interface_init_masks
   END INTERFACE module_interface_init_masks

   INTERFACE module_interface_define_netcdf_grid
      MODULE PROCEDURE module_interface_define_netcdf_grid
   END INTERFACE module_interface_define_netcdf_grid

   INTERFACE module_interface_init_arrays
      MODULE PROCEDURE module_interface_init_arrays
   END INTERFACE module_interface_init_arrays

   INTERFACE module_interface_init
      MODULE PROCEDURE module_interface_init
   END INTERFACE module_interface_init

   INTERFACE module_interface_header
      MODULE PROCEDURE module_interface_header
   END INTERFACE module_interface_header

   INTERFACE module_interface_actions
      MODULE PROCEDURE module_interface_actions
      MODULE PROCEDURE module_interface_actions_ij
   END INTERFACE module_interface_actions

   INTERFACE module_interface_swap_timelevel
      MODULE PROCEDURE module_interface_swap_timelevel
   END INTERFACE module_interface_swap_timelevel

   INTERFACE module_interface_3d_data_averaging
      MODULE PROCEDURE module_interface_3d_data_averaging
   END INTERFACE module_interface_3d_data_averaging

   INTERFACE module_interface_data_output_2d
      MODULE PROCEDURE module_interface_data_output_2d
   END INTERFACE module_interface_data_output_2d

   INTERFACE module_interface_data_output_3d
      MODULE PROCEDURE module_interface_data_output_3d
   END INTERFACE module_interface_data_output_3d

   INTERFACE module_interface_statistics
      MODULE PROCEDURE module_interface_statistics
   END INTERFACE module_interface_statistics

   INTERFACE module_interface_rrd_global
      MODULE PROCEDURE module_interface_rrd_global
   END INTERFACE module_interface_rrd_global

   INTERFACE module_interface_wrd_global
      MODULE PROCEDURE module_interface_wrd_global
   END INTERFACE module_interface_wrd_global

   INTERFACE module_interface_rrd_local
      MODULE PROCEDURE module_interface_rrd_local
   END INTERFACE module_interface_rrd_local

   INTERFACE module_interface_wrd_local
      MODULE PROCEDURE module_interface_wrd_local
   END INTERFACE module_interface_wrd_local

   INTERFACE module_interface_last_actions
      MODULE PROCEDURE module_interface_last_actions
   END INTERFACE module_interface_last_actions


CONTAINS


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Read module-specific parameter namelists
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_parin


   CALL bio_parin
   CALL bcm_parin
   CALL chem_parin
   CALL flight_parin ! ToDo: rename module to match filename
   CALL gust_parin
   CALL im_parin
   CALL lsm_parin
   ! ToDo: create parin routine for large_scale_forcing and nudging (should be seperate modules or new module switch)
   CALL mas_parin
   CALL nesting_offl_parin
   CALL ocean_parin
   CALL pcm_parin
   CALL radiation_parin
   CALL salsa_parin
   CALL spectra_parin
   CALL surface_data_output_parin
   CALL stg_parin
   CALL user_parin ! ToDo: make user code a single Fortran module
   CALL usm_parin
   CALL vm_parin
   CALL wtm_parin

   CALL package_parin ! ToDo: deprecated, needs to be dissolved


END SUBROUTINE module_interface_parin


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Perform module-specific initialization checks
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_check_parameters


   IF ( biometeorology )       CALL bio_check_parameters
   IF ( bulk_cloud_model )     CALL bcm_check_parameters
   IF ( air_chemistry )        CALL chem_check_parameters
   IF ( gust_module_enabled )  CALL gust_check_parameters
   IF ( land_surface )         CALL lsm_check_parameters
   IF ( large_scale_forcing  .OR.  nudging )  CALL lsf_nudging_check_parameters ! ToDo: create single module switch
   IF ( nesting_offline )      CALL nesting_offl_check_parameters
   IF ( ocean_mode )           CALL ocean_check_parameters
   IF ( plant_canopy )         CALL pcm_check_parameters
   IF ( radiation )            CALL radiation_check_parameters
   IF ( salsa )                CALL salsa_check_parameters
   IF ( calculate_spectra )    CALL spectra_check_parameters
   IF ( surface_output )       CALL surface_data_output_check_parameters
   IF ( syn_turb_gen )         CALL stg_check_parameters
   IF ( urban_surface )        CALL usm_check_parameters
   IF ( wind_turbine )         CALL wtm_check_parameters
                               CALL user_check_parameters


END SUBROUTINE module_interface_check_parameters


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Check module-specific data output of profiles
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_check_data_output_pr( variable, var_count, unit,   &
                                                dopr_unit )


   CHARACTER (LEN=*), INTENT(IN)    ::  variable  !< variable name
   INTEGER(iwp),      INTENT(IN)    ::  var_count !< variable output array index
   CHARACTER (LEN=*), INTENT(INOUT) ::  unit      !< physical unit of variable
   CHARACTER (LEN=*), INTENT(OUT)   ::  dopr_unit !< local value of dopr_unit

   IF ( unit == 'illegal' .AND.  bulk_cloud_model )  THEN
      CALL bcm_check_data_output_pr( variable, var_count, unit, dopr_unit )
   ENDIF

   IF ( unit == 'illegal' .AND.  air_chemistry )  THEN
      CALL chem_check_data_output_pr( variable, var_count, unit, dopr_unit )
   ENDIF

   IF ( unit == 'illegal'  .AND.  gust_module_enabled  )  THEN
      CALL gust_check_data_output_pr( variable, var_count, unit, dopr_unit )
   ENDIF

   IF ( unit == 'illegal' )  THEN ! ToDo: add module switch if possible
      CALL lsm_check_data_output_pr( variable, var_count, unit, dopr_unit )
   ENDIF

   IF ( unit == 'illegal' )  THEN ! ToDo: add module switch if possible
      CALL lsf_nudging_check_data_output_pr( variable, var_count, unit, dopr_unit )
   ENDIF

   IF ( unit == 'illegal'  .AND.  ocean_mode )  THEN
      CALL ocean_check_data_output_pr( variable, var_count, unit, dopr_unit )
   ENDIF

   IF ( unit == 'illegal'  .AND.  radiation )  THEN
      CALL radiation_check_data_output_pr( variable, var_count, unit, dopr_unit )
   ENDIF

   IF ( unit == 'illegal' )  THEN
      unit = '' ! ToDo: Seems like a hack. Find a general soultion!
      CALL user_check_data_output_pr( variable, var_count, unit )
   ENDIF



END SUBROUTINE module_interface_check_data_output_pr

!------------------------------------------------------------------------------!
! Description:
! ------------
!> Check module-specific 2D and 3D data output
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_check_data_output( variable, unit, i, ilen, k )


   CHARACTER (LEN=*), INTENT(IN)    ::  variable !< variable name
   CHARACTER (LEN=*), INTENT(INOUT) ::  unit     !< physical unit of variable

   INTEGER(iwp),      INTENT(IN)    :: i         !< ToDo: remove dummy argument, instead pass string from data_output
   INTEGER(iwp),      INTENT(IN)    :: ilen      !< ToDo: remove dummy argument, instead pass string from data_output
   INTEGER(iwp),      INTENT(IN)    :: k         !< ToDo: remove dummy argument, instead pass string from data_output

   IF ( unit == 'illegal'  .AND.  biometeorology )  THEN
      CALL bio_check_data_output( variable, unit, i, ilen, k )
   ENDIF

   IF ( unit == 'illegal'  .AND.  bulk_cloud_model  )  THEN
      CALL bcm_check_data_output( variable, unit )
   ENDIF

   IF ( unit == 'illegal'  .AND.  air_chemistry                                &
        .AND.  (variable(1:3) == 'kc_' .OR. variable(1:3) == 'em_') )  THEN  ! ToDo: remove aditional conditions
      CALL chem_check_data_output( variable, unit, i, ilen, k )
   ENDIF

   IF ( unit == 'illegal'  .AND.  gust_module_enabled  )  THEN
      CALL gust_check_data_output( variable, unit )
   ENDIF

   IF ( unit == 'illegal' )  THEN  ! ToDo: add module switch if possible
      CALL lsm_check_data_output( variable, unit, i, ilen, k )
   ENDIF

   IF ( unit == 'illegal'  .AND.  ocean_mode )  THEN
      CALL ocean_check_data_output( variable, unit )
   ENDIF

   IF ( unit == 'illegal'  .AND.  plant_canopy                                 &
        .AND.  variable(1:4) == 'pcm_' )  THEN  ! ToDo: remove aditional conditions
      CALL pcm_check_data_output( variable, unit )
   ENDIF

   IF ( unit == 'illegal'  .AND.  radiation )  THEN
      CALL radiation_check_data_output( variable, unit, i, ilen, k )
   ENDIF

   IF ( unit == 'illegal' .AND. salsa ) THEN
       CALL salsa_check_data_output( variable, unit )
   ENDIF

   IF ( unit == 'illegal'  .AND.  urban_surface                      &
        .AND.  variable(1:4) == 'usm_' )  THEN  ! ToDo: remove aditional conditions
       CALL usm_check_data_output( variable, unit )
   ENDIF

   IF ( unit == 'illegal' )  THEN
      unit = ''
      CALL user_check_data_output( variable, unit )
   ENDIF


END SUBROUTINE module_interface_check_data_output


!------------------------------------------------------------------------------!
!
! Description:
! ------------
!> Interface for init_masks. ToDo: get rid of these redundant calls!
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_init_masks( variable, unit )


   CHARACTER (LEN=*), INTENT(IN)    ::  variable !< variable name
   CHARACTER (LEN=*), INTENT(INOUT) ::  unit     !< physical unit of variable

   IF ( unit == 'illegal'  .AND.  air_chemistry                                &
        .AND.  (variable(1:3) == 'kc_' .OR. variable(1:3) == 'em_') )  THEN  ! ToDo: remove aditional conditions
      CALL chem_check_data_output( variable, unit, 0, 0, 0 )
   ENDIF

   IF ( unit == 'illegal'  .AND.  radiation )  THEN
      CALL radiation_check_data_output( variable, unit, 0, 0, 0 )
   ENDIF

   IF ( unit == 'illegal'  .AND.  salsa )  THEN
      CALL salsa_check_data_output( variable, unit )
   ENDIF

   IF ( unit == 'illegal' )  THEN
      unit = ''
      CALL user_check_data_output( variable, unit )
   ENDIF


END SUBROUTINE module_interface_init_masks


!------------------------------------------------------------------------------!
!
! Description:
! ------------
!> Define appropriate grid for module-specific netcdf output variables.
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_define_netcdf_grid( var, found,                    &
                                                grid_x, grid_y, grid_z )


   CHARACTER (LEN=*), INTENT(IN)  ::  var    !< variable name
   LOGICAL,           INTENT(OUT) ::  found  !< indicates if variable was found
   CHARACTER (LEN=*), INTENT(OUT) ::  grid_x !< netcdf dimension in x-direction
   CHARACTER (LEN=*), INTENT(OUT) ::  grid_y !< netcdf dimension in y-direction
   CHARACTER (LEN=*), INTENT(OUT) ::  grid_z !< netcdf dimension in z-direction
!
!--As long as no action is done in this subroutine, initialize strings with   
!--intent(out) attribute, in order to avoid compiler warnings.
   found  = .FALSE.
   grid_x = 'none'
   grid_y = 'none'
   grid_z = 'none'

END SUBROUTINE module_interface_define_netcdf_grid


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Allocate module-specific arrays and pointers
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_init_arrays


   IF ( bulk_cloud_model    )  CALL bcm_init_arrays
   IF ( air_chemistry       )  CALL chem_init_arrays
   IF ( gust_module_enabled )  CALL gust_init_arrays
   IF ( land_surface        )  CALL lsm_init_arrays
   IF ( ocean_mode          )  CALL ocean_init_arrays
   IF ( salsa               )  CALL salsa_init_arrays
   IF ( urban_surface       )  CALL usm_init_arrays
   IF ( wind_turbine        )  CALL wtm_init_arrays


END SUBROUTINE module_interface_init_arrays


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Perform module-specific initialization
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_init


   IF ( biometeorology      )  CALL bio_init
   IF ( bulk_cloud_model    )  CALL bcm_init
   IF ( air_chemistry       )  CALL chem_init
   IF ( virtual_flight      )  CALL flight_init
   IF ( gust_module_enabled )  CALL gust_init
   IF ( indoor_model        )  CALL im_init
   IF ( large_scale_forcing )  CALL lsf_init
   IF ( land_surface        )  CALL lsm_init
   IF ( nudging             )  CALL nudge_init
   IF ( ocean_mode          )  CALL ocean_init
   IF ( plant_canopy        )  CALL pcm_init
   IF ( salsa               )  CALL salsa_init
   IF ( urban_surface       )  CALL usm_init
   IF ( virtual_measurement )  CALL vm_init
   IF ( wind_turbine        )  CALL wtm_init
   IF ( radiation           )  CALL radiation_init
                               CALL user_init


END SUBROUTINE module_interface_init


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Gather module-specific header output
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_header( io )


   INTEGER(iwp), INTENT(IN) ::  io  !< unit of the output file


   IF ( biometeorology      )  CALL bio_header ( io )
   IF ( bulk_cloud_model    )  CALL bcm_header( io )
   IF ( air_chemistry       )  CALL chem_header ( io )
   IF ( virtual_flight      )  CALL flight_header( io )
   IF ( gust_module_enabled )  CALL gust_header( io )
   IF ( land_surface        )  CALL lsm_header( io )
   IF ( large_scale_forcing )  CALL lsf_nudging_header( io )
   IF ( nesting_offline     )  CALL nesting_offl_header( io )
   IF ( ocean_mode          )  CALL ocean_header( io )
   IF ( plant_canopy        )  CALL pcm_header( io )
   IF ( radiation           )  CALL radiation_header( io )
   IF ( salsa               )  CALL salsa_header( io )
   IF ( calculate_spectra   )  CALL spectra_header( io )
   IF ( syn_turb_gen        )  CALL stg_header( io )

   CALL user_header( io ) ! ToDo: make user module switch


END SUBROUTINE module_interface_header


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Perform module-specific actions while in time-integration (cache-optimized)
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_actions( location )


   CHARACTER (LEN=*), INTENT(IN) ::  location !< call location string


   IF ( gust_module_enabled )  CALL gust_actions( location )

   CALL user_actions( location ) ! ToDo: make user module switch


END SUBROUTINE module_interface_actions


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Perform module-specific actions while in time-integration (vector-optimized)
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_actions_ij( i, j, location )


   INTEGER(iwp),      INTENT(IN) ::  i         !< grid index in x-direction
   INTEGER(iwp),      INTENT(IN) ::  j         !< grid index in y-direction
   CHARACTER (LEN=*), INTENT(IN) ::  location  !< call location string


   IF ( gust_module_enabled )  CALL gust_actions( i, j, location )

   CALL user_actions( i, j, location ) ! ToDo: make user module switch


END SUBROUTINE module_interface_actions_ij


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Swap the timelevel pointers for module-specific arrays
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_swap_timelevel ( swap_mode )


   INTEGER(iwp), INTENT(IN) :: swap_mode !< determines procedure of pointer swap

   IF ( bulk_cloud_model    )  CALL bcm_swap_timelevel( swap_mode )
   IF ( air_chemistry       )  CALL chem_swap_timelevel( swap_mode )
   IF ( gust_module_enabled )  CALL gust_swap_timelevel( swap_mode )
   IF ( land_surface        )  CALL lsm_swap_timelevel( swap_mode )
   IF ( ocean_mode          )  CALL ocean_swap_timelevel( swap_mode )
   IF ( salsa               )  CALL salsa_swap_timelevel( swap_mode )
   IF ( urban_surface       )  CALL usm_swap_timelevel( swap_mode )


END SUBROUTINE module_interface_swap_timelevel


!------------------------------------------------------------------------------!
!
! Description:
! ------------
!> Perform module-specific averaging of 3D data
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_3d_data_averaging( mode, variable )


   CHARACTER (LEN=*), INTENT(IN) ::  mode     !< averaging interface mode
   CHARACTER (LEN=*), INTENT(IN) ::  variable !< variable name

   IF ( biometeorology      )  CALL bio_3d_data_averaging( mode, variable )
   IF ( bulk_cloud_model    )  CALL bcm_3d_data_averaging( mode, variable )
   IF ( air_chemistry       )  CALL chem_3d_data_averaging( mode, variable )
   IF ( gust_module_enabled )  CALL gust_3d_data_averaging( mode, variable )
   IF ( land_surface        )  CALL lsm_3d_data_averaging( mode, variable )
   IF ( ocean_mode          )  CALL ocean_3d_data_averaging( mode, variable )
   IF ( radiation           )  CALL radiation_3d_data_averaging( mode, variable )
   IF ( salsa               )  CALL salsa_3d_data_averaging( mode, variable )
   IF ( urban_surface       )  CALL usm_3d_data_averaging( mode, variable )

   CALL user_3d_data_averaging( mode, variable ) ! ToDo: make user module switch


END SUBROUTINE module_interface_3d_data_averaging

!------------------------------------------------------------------------------!
!
! Description:
! ------------
!> Define module-specific 2D output variables
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_data_output_2d( av, variable, found, grid, mode,   &
                                            local_pf, two_d, nzb_do, nzt_do,   &
                                            fill_value )

   INTEGER(iwp),      INTENT(IN)    ::  av         !< flag for (non-)average output
   CHARACTER (LEN=*), INTENT(IN)    ::  variable   !< variable name
   LOGICAL,           INTENT(INOUT) ::  found      !< flag if output variable is found
   CHARACTER (LEN=*), INTENT(INOUT) ::  grid       !< name of vertical grid
   CHARACTER (LEN=*), INTENT(IN)    ::  mode       !< either 'xy', 'xz' or 'yz'
   LOGICAL,           INTENT(OUT)   ::  two_d      !< flag for 2D variables
   INTEGER(iwp),      INTENT(IN)    ::  nzb_do     !< vertical output index (bottom) (usually 0)
   INTEGER(iwp),      INTENT(IN)    ::  nzt_do     !< vertical output index (top) (usually nz_do3d)
   REAL(wp),          INTENT(IN)    ::  fill_value !< to be removed

   REAL(wp), DIMENSION(nxl:nxr,nys:nyn,nzb_do:nzt_do), INTENT(OUT) ::  local_pf !< ToDo: can also be kind=sp


   IF ( .NOT. found  .AND.  biometeorology )  THEN
      CALL bio_data_output_2d(                                                 &
              av, variable, found, grid, local_pf, two_d, nzb_do, nzt_do       &
           )
   ENDIF

   IF ( .NOT. found  .AND.  bulk_cloud_model )  THEN
      CALL bcm_data_output_2d(                                                 &
              av, variable, found, grid, mode, local_pf, two_d, nzb_do, nzt_do &
           )
   ENDIF

   IF ( .NOT. found  .AND.  air_chemistry )  THEN
      CALL chem_data_output_2d(                                                &
              av, variable, found, grid, mode, local_pf, two_d, nzb_do, nzt_do, fill_value &
           )
   ENDIF

   IF ( .NOT. found  .AND.  gust_module_enabled )  THEN
      CALL gust_data_output_2d(                                                &
              av, variable, found, grid, mode, local_pf, two_d, nzb_do, nzt_do, fill_value &
           )
   ENDIF

   IF ( .NOT. found  .AND.  land_surface )  THEN
      CALL lsm_data_output_2d(                                                 &
              av, variable, found, grid, mode, local_pf, two_d, nzb_do, nzt_do &
           )
   ENDIF

   IF ( .NOT. found  .AND.  ocean_mode )  THEN
      CALL ocean_data_output_2d(                                               &
              av, variable, found, grid, mode, local_pf, nzb_do, nzt_do        &
           )
   ENDIF

   IF ( .NOT. found  .AND.  radiation )  THEN
      CALL radiation_data_output_2d(                                           &
              av, variable, found, grid, mode, local_pf, two_d, nzb_do, nzt_do &
           )
   ENDIF

   IF ( .NOT. found  .AND.  salsa )  THEN
      CALL salsa_data_output_2d(                                               &
              av, variable, found, grid, mode, local_pf, two_d, nzb_do, nzt_do &
           )
   ENDIF

   IF ( .NOT. found )  THEN  ! ToDo: make user module switch
      CALL user_data_output_2d(                                                &
              av, variable, found, grid, local_pf, two_d, nzb_do, nzt_do       &
           )
   ENDIF

END SUBROUTINE module_interface_data_output_2d


!------------------------------------------------------------------------------!
!
! Description:
! ------------
!> Define module-specific 3D output variables
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_data_output_3d( av, variable, found, local_pf,     &
                                            fill_value, resorted, nzb_do, nzt_do )


   INTEGER(iwp),      INTENT(IN)    ::  av         !< flag for (non-)average output
   CHARACTER (LEN=*), INTENT(IN)    ::  variable   !< variable name
   LOGICAL,           INTENT(INOUT) ::  found      !< flag if output variable is found
   REAL(wp),          INTENT(IN)    ::  fill_value !< ToDo: refactor
   LOGICAL,           INTENT(OUT)   ::  resorted   !< flag if output has been resorted
   INTEGER(iwp),      INTENT(IN)    ::  nzb_do     !< vertical output index (bottom) (usually 0)
   INTEGER(iwp),      INTENT(IN)    ::  nzt_do     !< vertical output index (top) (usually nz_do3d)

   REAL(sp), DIMENSION(nxl:nxr,nys:nyn,nzb_do:nzt_do), INTENT(OUT) ::  local_pf


   IF ( .NOT. found  .AND.  biometeorology )  THEN
      CALL bio_data_output_3d( av, variable, found, local_pf, nzb_do, nzt_do )
      resorted = .FALSE.
   ENDIF

   IF ( .NOT. found  .AND.  bulk_cloud_model )  THEN
      CALL bcm_data_output_3d( av, variable, found, local_pf, nzb_do, nzt_do )
      resorted = .TRUE.
   ENDIF

   IF ( .NOT. found  .AND.  air_chemistry )  THEN
      CALL chem_data_output_3d( av, variable, found, local_pf, fill_value, nzb_do, nzt_do )
      resorted = .TRUE.
   ENDIF

   IF ( .NOT. found  .AND.  gust_module_enabled )  THEN
      CALL gust_data_output_3d( av, variable, found, local_pf, fill_value, nzb_do, nzt_do )
      resorted = .TRUE.
   ENDIF

   IF ( .NOT. found  .AND.  ocean_mode )  THEN
      CALL ocean_data_output_3d( av, variable, found, local_pf, nzb_do, nzt_do )
      resorted = .TRUE.
   ENDIF

   IF ( .NOT. found  .AND.  plant_canopy )  THEN
      CALL pcm_data_output_3d( av, variable, found, local_pf, fill_value, nzb_do, nzt_do )
      resorted = .TRUE.
   ENDIF

   IF ( .NOT. found  .AND.  radiation )  THEN
      CALL radiation_data_output_3d( av, variable, found, local_pf, nzb_do, nzt_do )
      resorted = .TRUE.
   ENDIF

   IF ( .NOT. found  .AND.  salsa )  THEN
      CALL salsa_data_output_3d( av, variable, found, local_pf, nzb_do, nzt_do )
      resorted = .TRUE.
   ENDIF

   IF ( .NOT. found )  THEN  ! ToDo: make user module switch
      CALL user_data_output_3d( av, variable, found, local_pf, nzb_do, nzt_do )
      resorted = .TRUE.
   ENDIF


END SUBROUTINE module_interface_data_output_3d


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Compute module-specific profile and timeseries data
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_statistics( mode, sr, tn, dots_max )


   CHARACTER (LEN=*), INTENT(IN) ::  mode     !< statistical analysis mode
   INTEGER(iwp),      INTENT(IN) ::  sr       !<
   INTEGER(iwp),      INTENT(IN) ::  tn       !<
   INTEGER(iwp),      INTENT(IN) ::  dots_max !< maximum number of timeseries

   IF ( gust_module_enabled )  CALL gust_statistics( mode, sr, tn, dots_max )
   IF ( air_chemistry       )  CALL chem_statistics( mode, sr, tn )

   CALL user_statistics( mode, sr, tn ) ! ToDo: make user module switch


END SUBROUTINE module_interface_statistics


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Read module-specific restart data globaly shared by all MPI ranks
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_rrd_global( variable, found )


   CHARACTER (LEN=*), INTENT(IN)  ::  variable !< name of global variable
   LOGICAL,           INTENT(OUT) ::  found    !< flag if variable was found

   IF ( .NOT. found )  CALL bio_rrd_global( found ) ! ToDo: change interface to pass variable
   IF ( .NOT. found )  CALL bcm_rrd_global( found ) ! ToDo: change interface to pass variable
   IF ( .NOT. found )  CALL flight_rrd_global( found ) ! ToDo: change interface to pass variable
   IF ( .NOT. found )  CALL gust_rrd_global( found ) ! ToDo: change interface to pass variable
   IF ( .NOT. found )  CALL ocean_rrd_global( found ) ! ToDo: change interface to pass variable
   IF ( .NOT. found )  CALL stg_rrd_global ( found ) ! ToDo: change interface to pass variable
   IF ( .NOT. found )  CALL wtm_rrd_global( found ) ! ToDo: change interface to pass variable

   IF ( .NOT. found )  CALL user_rrd_global( found ) ! ToDo: change interface to pass variable


END SUBROUTINE module_interface_rrd_global


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Write module-specific restart data globaly shared by all MPI ranks
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_wrd_global


   IF ( biometeorology )       CALL bio_wrd_global
   IF ( bulk_cloud_model )     CALL bcm_wrd_global
   IF ( virtual_flight )       CALL flight_wrd_global
   IF ( gust_module_enabled )  CALL gust_wrd_global
   IF ( ocean_mode )           CALL ocean_wrd_global
   IF ( syn_turb_gen )         CALL stg_wrd_global
   IF ( wind_turbine )         CALL wtm_wrd_global

   CALL user_wrd_global ! ToDo: make user module switch


END SUBROUTINE module_interface_wrd_global


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Read module-specific restart data specific to local MPI ranks
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_rrd_local( variable, file_index, map_index,        &
                                       nxlf, nxlc, nxl_on_file,                &
                                       nxrf, nxrc, nxr_on_file,                &
                                       nynf, nync, nyn_on_file,                &
                                       nysf, nysc, nys_on_file,                &
                                       tmp_2d, tmp_3d, found )


   CHARACTER (LEN=*), INTENT(IN)  ::  variable     !< name of global variable
   INTEGER(iwp),      INTENT(IN)  ::  file_index   !<
   INTEGER(iwp),      INTENT(IN)  ::  map_index    !<
   INTEGER(iwp),      INTENT(IN)  ::  nxlc         !<
   INTEGER(iwp),      INTENT(IN)  ::  nxlf         !<
   INTEGER(iwp),      INTENT(IN)  ::  nxl_on_file  !<
   INTEGER(iwp),      INTENT(IN)  ::  nxrc         !<
   INTEGER(iwp),      INTENT(IN)  ::  nxrf         !<
   INTEGER(iwp),      INTENT(IN)  ::  nxr_on_file  !<
   INTEGER(iwp),      INTENT(IN)  ::  nync         !<
   INTEGER(iwp),      INTENT(IN)  ::  nynf         !<
   INTEGER(iwp),      INTENT(IN)  ::  nyn_on_file  !<
   INTEGER(iwp),      INTENT(IN)  ::  nysc         !<
   INTEGER(iwp),      INTENT(IN)  ::  nysf         !<
   INTEGER(iwp),      INTENT(IN)  ::  nys_on_file  !<
   LOGICAL,           INTENT(OUT) ::  found        !< flag if variable was found

   REAL(wp), DIMENSION(nys_on_file-nbgp:nyn_on_file+nbgp,nxl_on_file-nbgp:nxr_on_file+nbgp), INTENT(OUT) :: tmp_2d   !<
   REAL(wp), DIMENSION(nzb:nzt+1,nys_on_file-nbgp:nyn_on_file+nbgp,nxl_on_file-nbgp:nxr_on_file+nbgp), INTENT(OUT) :: tmp_3d   !<


   IF ( .NOT. found ) CALL bio_rrd_local(                                      &
                              found                                            &
                           )

   IF ( .NOT. found ) CALL bcm_rrd_local(                                      &
                              file_index, map_index,                           &
                              nxlf, nxlc, nxl_on_file,                         &
                              nxrf, nxrc, nxr_on_file,                         &
                              nynf, nync, nyn_on_file,                         &
                              nysf, nysc, nys_on_file,                         &
                              tmp_2d, tmp_3d, found                            &
                           ) ! ToDo: change interface to pass variable

   IF ( .NOT. found ) CALL chem_rrd_local(                                     &
                              file_index, map_index,                           &
                              nxlf, nxlc, nxl_on_file,                         &
                              nxrf, nxrc, nxr_on_file,                         &
                              nynf, nync, nyn_on_file,                         &
                              nysf, nysc, nys_on_file,                         &
                              tmp_3d, found                                    &
                           ) ! ToDo: change interface to pass variable

   IF ( .NOT. found ) CALL gust_rrd_local(                                     &
                              file_index, map_index,                           &
                              nxlf, nxlc, nxl_on_file,                         &
                              nxrf, nxrc, nxr_on_file,                         &
                              nynf, nync, nyn_on_file,                         &
                              nysf, nysc, nys_on_file,                         &
                              tmp_2d, tmp_3d, found                            &
                           ) ! ToDo: change interface to pass variable

   IF ( .NOT. found ) CALL lsm_rrd_local(                                      &
                              file_index, map_index,                           &
                              nxlf, nxlc, nxl_on_file,                         &
                              nxrf, nxrc, nxr_on_file,                         &
                              nynf, nync, nyn_on_file,                         &
                              nysf, nysc, nys_on_file,                         &
                              tmp_2d, found                                    &
                           ) ! ToDo: change interface to pass variable

   IF ( .NOT. found ) CALL ocean_rrd_local(                                    &
                              file_index, map_index,                           &
                              nxlf, nxlc, nxl_on_file,                         &
                              nxrf, nxrc, nxr_on_file,                         &
                              nynf, nync, nyn_on_file,                         &
                              nysf, nysc, nys_on_file,                         &
                              tmp_2d, tmp_3d, found                            &
                           ) ! ToDo: change interface to pass variable

   IF ( .NOT. found ) CALL radiation_rrd_local(                                &
                              file_index, map_index,                           &
                              nxlf, nxlc, nxl_on_file,                         &
                              nxrf, nxrc, nxr_on_file,                         &
                              nynf, nync, nyn_on_file,                         &
                              nysf, nysc, nys_on_file,                         &
                              tmp_2d, tmp_3d, found                            &
                           ) ! ToDo: change interface to pass variable

   IF ( .NOT. found ) CALL salsa_rrd_local(                                    &
                              file_index, map_index,                           &
                              nxlf, nxlc, nxl_on_file,                         &
                              nxrf, nxrc, nxr_on_file,                         &
                              nynf, nync, nyn_on_file,                         &
                              nysf, nysc, nys_on_file,                         &
                              tmp_3d, found                                    &
                           ) ! ToDo: change interface to pass variable

   IF ( .NOT. found ) CALL usm_rrd_local(                                      &
                              file_index, map_index,                           &
                              nxlf, nxlc, nxl_on_file,                         &
                              nxrf, nxrc, nxr_on_file,                         &
                              nynf, nync, nyn_on_file,                         &
                              nysf, nysc, nys_on_file,                         &
                              found                                            &
                           ) ! ToDo: change interface to pass variable

   IF ( .NOT. found ) CALL user_rrd_local(                                     &
                              file_index, map_index,                           &
                              nxlf, nxlc, nxl_on_file,                         &
                              nxrf, nxrc, nxr_on_file,                         &
                              nynf, nync, nyn_on_file,                         &
                              nysf, nysc, nys_on_file,                         &
                              tmp_3d, found                                    &
                           ) ! ToDo: change interface to pass variable


END SUBROUTINE module_interface_rrd_local


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Write module-specific restart data specific to local MPI ranks
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_wrd_local


   IF ( biometeorology )       CALL bio_wrd_local
   IF ( bulk_cloud_model )     CALL bcm_wrd_local
   IF ( air_chemistry )        CALL chem_wrd_local
   IF ( gust_module_enabled )  CALL gust_wrd_local
   IF ( land_surface )         CALL lsm_wrd_local
   IF ( ocean_mode )           CALL ocean_wrd_local
   IF ( radiation )            CALL radiation_wrd_local
   IF ( salsa )                CALL salsa_wrd_local
   IF ( urban_surface )        CALL usm_wrd_local

   CALL user_wrd_local ! ToDo: make user module switch


END SUBROUTINE module_interface_wrd_local


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Perform module-specific last actions before the program terminates
!------------------------------------------------------------------------------!
SUBROUTINE module_interface_last_actions


   CALL user_last_actions ! ToDo: make user module switch


END SUBROUTINE module_interface_last_actions


END MODULE module_interface
